/* $Id: 542x.c,v 1.10 1998/11/24 18:28:24 marcus Exp $
******************************************************************************

   CL-GD5452x ramdac driver

   Copyright (C) 1995-1996 Steffen Seeger	[seeger@ggi-project.org]
   Copyright (C) 1997      Jason McMullan	[jmcc@ggi-project.org]

   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

******************************************************************************
*/

#include <kgi/maintainers.h>
#define	MAINTAINER	Martin_Erhardsen
#define DRIVER_NAME	"Cirrus Logic 542x DAC driver"
#define DRIVER_REV	"$Revision: 1.10 $"

#include <kgi/kgi.h>
#include <kgi/module.h>
#include <kgi/kgi_commands.h>


static void dac_set_clut(struct kgi_display *dpy, kgi_clut *clut,
			kgi_uint first, kgi_uint cnt)
{
	kgiu16 *data = (kgiu16 *) (clut->data + first);
	ASSERT(cnt);
	cnt *= 3;
	outb(first, 0x3C8);

	while (cnt--) outb(*(data++) >> 10, 0x3C9);
}

static void dac_get_clut(struct kgi_display *dpy,
			 kgi_clut *clut, kgi_uint first, int cnt)
{
	kgiu16 *data = (kgiu16 *) (clut->data + first);
	cnt *= 3;
	outb(first, 0x3C7);

	while (cnt--) *(data++) = inb(0x3C9)*1040;
}

/* This structure is initialized by the chipset driver */

struct kgi_ramdac 
ramdac = {	"Cirrus",	       		/* manufacturer		*/
	 	"542x",				/* model		*/
 		GM_ALL,				/* supported modes	*/
	 	0,			 	/* flags		*/
	 	{ 0xFFFF, 0xFFFF },		/* maximum resolution	*/
		NULL,				/* Private data hook	*/
		 DAC_16BIT_PB, 			/* device type		*/
		  {       0,  85226000 },	/* normal mode		*/
		  {       0,         0 },	/* 2xclk mode		*/

		 {{	  0,  32000000 },	/* 	KGIGT_TEXT16	*/
		  {	  0,         0 },	/* 	KGIGT_TEXT32 	*/
		  {       0,  85226000 },	/* 	KGIGT_1BIT		*/
		  {	  0,  85226000 },	/* 	KGIGT_4BIT		*/
		  {       0,  85226000 },	/* 	KGIGT_8BIT		*/
		  {	  0,  45076000 },	/* 	KGIGT_15BIT	*/
		  { 	  0,  45076000 },	/* 	KGIGT_16BIT	*/
		  {       0,  30000000 },	/* 	KGIGT_24BIT	*/
		  {	  0,  22538000 }}	/* 	KGIGT_32BIT	*/
         };

/*
	GGI interface
*/
int kgim_ramdac_init(struct kgi_card *card, struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " rev. " DRIVER_REV);
	card->ramdac=&ramdac;
	return EOK;
}

void kgim_ramdac_done(struct kgi_display *dpy)
{
	NOTICE(DRIVER_NAME " removed.");
}

int
kgim_ramdac_check_mode (struct kgi_display *dpy,
                        struct kgi_mode *mode, enum kgi_tm_cmd *cmd)
{
  int newclock = mode->dclk;
  
  if (!((1<<mode->request.graphtype)&ramdac.modes))
   {
  	DEBUG2("couldn't handle graphtype %i",mode->request.graphtype);
 	return -E(RAMDAC,MODE_CHECK);
   }

  switch (*cmd)
   {
     case CMD_PROPOSE:
	mode->lclk.mul = mode->clk.mul = 1;
        mode->lclk.div = mode->clk.div = 1;
	return EOK;

     case CMD_LOWER:
	if (   (newclock < ramdac.clock[mode->request.graphtype].min)
	    || (newclock < ramdac.normal.min))
	  {
	    DEBUG2("DCLK too low (%i Hz).", newclock);
	    return -E(RAMDAC,CLK_LOW_LIMIT);
	  }

	if (ramdac.clock[mode->request.graphtype].max < newclock)
    	  newclock = ramdac.clock[mode->request.graphtype].max;

	mode->dclk   = newclock;
	return EOK;

     case CMD_RAISE:
	if (   (newclock > ramdac.clock[mode->request.graphtype].max)
            || (newclock > ramdac.normal.max))
	  {
	    DEBUG2("DCLK too high (%i Hz).", newclock);
	    return -E(RAMDAC,CLK_HIGH_LIMIT);
	  }

	if (newclock < ramdac.clock[mode->request.graphtype].min)
	  newclock = ramdac.clock[mode->request.graphtype].min;

	if (newclock < ramdac.normal.min)
	  newclock = ramdac.normal.min;

	mode->dclk = newclock;
	return EOK;

     case CMD_CHECK:
        if ((mode->lclk.mul != 1) || (mode->lclk.div != 1)
	    /* || (mode->clk.mul  != 1) */ || (mode->clk.div  != 1))
	  return -E(RAMDAC,MODE_CHECK);

	if (   (mode->dclk < ramdac.normal.min)
	    || (mode->dclk < ramdac.clock[mode->request.graphtype].min))
	  return -E(RAMDAC,CLK_LOW_LIMIT);

	if (   (mode->dclk > ramdac.normal.max)
	    || (mode->dclk > ramdac.clock[mode->request.graphtype].max))
	  return -E(RAMDAC,CLK_HIGH_LIMIT);

	return EOK;

    default:
	INTERNAL_ERROR;
	return -E(RAMDAC,BOGUS);
  }
}

void kgim_ramdac_set_mode(struct kgi_display *dpy,struct kgi_mode *mode)
{
	outb(0xFF, 0x3C6);
        dpy->set_clut = dac_set_clut;
}

int kgim_ramdac_command(struct kgi_graphic *graph, kgi_uint cmd, kgi_uint arg)
{
	/*	this may set 
	**	-  dac_params (take care if 504 or 505 was detected).
	**	-  hardware cursor
	**	-  CLUT (or gamma correction table).
	*/
	kgi_clut clut;

	switch(cmd) { 
		case RAMDAC_GETINFO:
			graph->io_res=&ramdac;
			return EOK;
		case RAMDAC_SETINFO:
			if (suser()) {
				memcpy(&ramdac,graph->io_buf,sizeof(ramdac));
				return EOK;
			} else  return -EPERM;
		case RAMDAC_SETCLUT:
			{  clut.size=256;
			   clut.data=(kgi_color *)(graph->io_buf);
			   dac_set_clut(graph->dev.dpy,&clut,0,256);
			} 
			return EOK;
		case RAMDAC_GETCLUT:
			{  clut.size=256;
			   clut.data=(kgi_color *)(graph->io_buf);
			   graph->io_res=graph->io_buf;
			   dac_get_clut(graph->dev.dpy,&clut,0,256);
			} 
			return EOK;
	}
  	return -E(RAMDAC,NOSUP_ALWAYS_CANT);  /* This driver can't do this. */
}






